
/*
  **
  ** dns.c
  **
  ** Part of the drpoxy package by Matthew Pratt. 
  **
  ** Copyright 2000 Andreas Hofmeister (hofmeist@informatik.uni-freiburg.de)
  **
  ** This software is licensed under the terms of the GNU General 
  ** Public License (GPL). Please see the file COPYING for details.
  ** 
  **
*/

#include <config.h>

#include "conf.h"
#include "tmdns.h"
#include "dns.h"
#include "debug.h"

#define BYTESLEFT(x) (x->bufsize - (x->data - x->u.raw) + sizeof(HEADER))

static dns_soa_rr * fake_soa = NULL;
static char servername[MAXDNAME];

/****************************************************************************
 * get the length of the dns message.
 * 
 * @param pkt   a pointer to a dns package structure.
 ****************************************************************************/
int dns_get_len( dns_t * pkt ) {
	return (pkt->data - pkt->u.raw);
}
/****************************************************************************
 * init a dns packet structure.
 * 
 * @param pkt   a pointer to a dns package structure.
 ****************************************************************************/
void dns_init( dns_t * pkg) {

    memset( pkg , 0 , sizeof(dns_t) );
    pkg->bufsize    = DNSDATASIZE;
    pkg->data       = pkg->u.raw + sizeof(HEADER);
    pkg->dnptrs[0]  = pkg->data;
    pkg->last_dnptr = pkg->dnptrs[MAXDOMAINS];

}
/****************************************************************************
 * init a dns answer.
 * 
 * @param pkg   a pointer to a dns package structure.
 ****************************************************************************/
int dns_init_answer( dns_t * pkt, dns_query * query ) {

  /** initialize header fields */
  if( query && query->inverse ) {
      pkt->u.hdr.opcode = IQUERY;
  } else {
      pkt->u.hdr.opcode = QUERY;
  }
  
  if( query ) {
      pkt->u.hdr.id     = query->id;
  } else {
      pkt->u.hdr.id     = rand();
  }

  pkt->u.hdr.qr     = 1;       /* this is a response                   */
  pkt->u.hdr.aa     = 0;       /* never send authoritativ answers      */
  pkt->u.hdr.tc     = 0;       /* not truncated, but we will see ...   */
  pkt->u.hdr.rd     = 0;       /* we always do recursion (when online) */
  pkt->u.hdr.ra     = 0;       /* recursion is not available           */
#if __BIND >= 19960801
  pkt->u.hdr.ad     = 0;       /* bind says 0 here, so we'll do        */
#endif
  pkt->u.hdr.rcode  = NOERROR; /* message is ok.                       */

  pkt->u.hdr.qdcount = 0;
  pkt->u.hdr.ancount = 0;
  pkt->u.hdr.nscount = 0;
  pkt->u.hdr.arcount = 0;

  if( query && query->want_recursion ) {
	pkt->u.hdr.rd = 1;
  };

  /** add the query to the packet */
  if( query && (! query->inverse) ) {
    int n = 0;
    int bytes = 0;

    for( n = 0; n < query->nquest; n ++ ) {
      bytes += dns_add_qr( pkt , 
		      	   query->question[n].query_arg , 
      			   query->question[n].query_type, 
			   query->question[n].query_class );
    }
    return bytes;
  }

  return 0;
}
/*****************************************************************************
 * compress a domain into a message
 *
 ****************************************************************************/
int dns_compress_domain( dns_t * pkt , char * domain ) {
    int count = 0;
    count = dn_comp( domain,
	             pkt->data   , BYTESLEFT(pkt) ,
		     //NULL, NULL );
		     pkt->dnptrs , &(pkt->last_dnptr) );
    if( count < 0 ) return -1;

    pkt->data += count;
    return count;
}
/*****************************************************************************
 * put a character string into a message
 *
 * character strings are at most 255 bytes long, all strings are
 * prefixed with a length byte.
 ****************************************************************************/
int dns_put_string( dns_t * pkt, char * string ) {

    int slen = strlen(string);

    if( slen > 255 ) slen = 255;
    if( BYTESLEFT(pkt) < slen + 1 ) return -1;

    *(pkt->data) = slen;
    pkt->data ++;
    memcpy( pkt->data , string , slen );
    pkt->data += slen;

    return slen + 1 ;
    
}
/*****************************************************************************
 * add a query record to a packet.
 *
 *  @param pkt     pointer to a dns packet structure.
 *  @param domian  domain for this RR
 *  @param type    type for this RR, currently T_A and T_PTR are supported.
 *  @param class   class for this RR, always set this to C_IN
 *
 *  @return length of this RR or -1 on error.
 ****************************************************************************/
int  dns_add_qr( dns_t * pkt , char * domain,
		 short type, short class )
{
    int count = 0;
    u_char * old_data = pkt->data;

    count = dns_compress_domain(pkt,domain);
    if( count < 0 ) return -1;

    if( BYTESLEFT(pkt) < 4 ) return -1;

    PUTSHORT( type   , pkt->data );
    PUTSHORT( class  , pkt->data );

    pkt->u.hdr.qdcount = htons( ntohs(pkt->u.hdr.qdcount) + 1);
    return pkt->data - old_data ;
}
/*****************************************************************************
 * add a new resource record to a packet.
 *
 *  @param pkt     pointer to a dns packet structure.
 *  @param rr      pointer to the data for this RR.
 *
 *  @return length of this RR or -1 on error.
 ****************************************************************************/
int dns_add_raw_rr( dns_t * pkt, char * domain, 
		short type , short class, long ttl ,
		int rr_len, u_char * rr_data ) 
{
    int count = 0;
    u_char * old_data = pkt->data;
    
    count = dns_compress_domain(pkt,domain);
    if( count < 0 ) return -1;
   
    if( BYTESLEFT(pkt) < 10 + rr_len ) return -1 ;

    PUTSHORT( type   , pkt->data );
    PUTSHORT( class  , pkt->data );
    PUTLONG ( ttl    , pkt->data );
    PUTSHORT( rr_len , pkt->data );
    memcpy(pkt->data,rr_data,rr_len);
    pkt->data += rr_len;

    pkt->u.hdr.ancount = htons( ntohs(pkt->u.hdr.ancount) + 1);
    return  pkt->data - old_data;
 
}
/*****************************************************************************
 * add a new resource record to a packet.
 *
 *  @param pkt     pointer to a dns packet structure.
 *  @param rr      pointer to the data for this RR.
 *
 *  @return length of this RR or -1 on error.
 ****************************************************************************/
int dns_add_rr( dns_t * pkt, dns_rr * rr )
{
    int count = 0;
    u_char * old_data = pkt->data;
    u_char * lenptr;

    
    count = dns_compress_domain(pkt,rr->domain);
    if( count < 0 ) return -1;
   
    if( BYTESLEFT(pkt) < 10 ) return -1 ;

    PUTSHORT( rr->type   , pkt->data );
    PUTSHORT( rr->class  , pkt->data );
    PUTLONG ( rr->ttl    , pkt->data );
    lenptr = pkt->data;
    PUTSHORT( 0 , pkt->data );

    switch( rr->type ) {
	case T_A:
	    if( BYTESLEFT(pkt) < sizeof(struct in_addr) ) return - 1;
	    memcpy( pkt->data , &(rr->rr.a) , sizeof(struct in_addr));
	    pkt->data += sizeof(struct in_addr);
    	    PUTSHORT( sizeof(struct in_addr) , lenptr);
	    break;
	case T_NS:
	case T_MD:
	case T_MF:
	case T_CNAME:
	case T_MB:
	case T_MG:
	case T_MR:
	case T_PTR:
	    /* this are the standard RR's which return a domain */
    	    count = dns_compress_domain(pkt,rr->rr.dn);
	    if( count < 0 ) {
	    	return -1;
	    }
    	    PUTSHORT( count , lenptr);
	    break;
	case T_SOA:
	    {
		short octets = 0;
		dns_soa_rr * soarec = & rr->rr.soa;
    	    	count = dns_compress_domain(pkt,soarec->name_server);
	    	if( count < 0 ) return -1;
		octets += count;
    	    	count = dns_compress_domain(pkt,soarec->admin );
	    	if( count < 0 ) return -1;
		octets += count;
		if( BYTESLEFT(pkt) < 5 * 4 ) return -1;
		PUTLONG(soarec->serial ,pkt->data);
		PUTLONG(soarec->refresh,pkt->data);
		PUTLONG(soarec->retry  ,pkt->data);
		PUTLONG(soarec->expire ,pkt->data);
		PUTLONG(soarec->min_ttl,pkt->data);
    	        PUTSHORT( octets += 5*4 , lenptr);
	    }
	    break;
	case T_MX:
	    {
		dns_mx_rr * mxrec = & rr->rr.mx ;
		PUTSHORT( mxrec->preference, pkt->data);
    	    	count = dns_compress_domain(pkt,mxrec->mail_host);
	    	if( count < 0 ) return -1;
    	        PUTSHORT( 2 + count , lenptr);
	    }
	    break;
	case T_HINFO:
	    {
		short datalen = 0;
		dns_hinfo_rr * hirec = & rr->rr.hinfo;
		count = dns_put_string(pkt,hirec->cpu_type);
		if( count < 0 ) return -1;
		datalen = count;
		
		count = dns_put_string(pkt,hirec->os_name);
		if( count < 0 ) return -1;
		datalen += count;

    	        PUTSHORT( datalen , lenptr);
	    }
	    break;
	case T_TXT:
	    {
		dns_txt_rr * txtrec = & rr->rr.txt;
		int idx = 0;
		int datalen = 0;

		if(txtrec->nstrings > DNSMAXTXT) 
		    txtrec->nstrings = DNSMAXTXT;

		for( idx = 0; idx < txtrec->nstrings; idx ++ ) {
		    if( txtrec->strings[idx] == NULL) break;
		    count = dns_put_string(pkt,txtrec->strings[idx]);
		    if( count < 0 ) return -1;
		    datalen = count;
		}
    	        PUTSHORT( datalen , lenptr);
	    }
	    break;
	case T_SRV:
	    {
		dns_srv_rr * srvrec = & rr->rr.srv;

    	        PUTSHORT( srvrec->priority, pkt->data );
    	        PUTSHORT( srvrec->weight,   pkt->data );
    	        PUTSHORT( srvrec->port,     pkt->data );
		//count = dns_put_string(pkt,srvrec->target);
    	    	count = dns_compress_domain(pkt,srvrec->target);
		if( count < 0 ) {
		    return -1;
		}
		debug("encoding SRV, strlen is %d , count is %d dlen is %d\n" , 
			strlen(srvrec->target) , count , count + 6 );
    	        PUTSHORT( count + 6 , lenptr);
	    }
	    break;
	default:
	    debug("Unknown rr type %d\n" , rr->type);
	    return -1;
    }
    
    pkt->u.hdr.ancount = htons( ntohs(pkt->u.hdr.ancount) + 1);
    return pkt->data - old_data ;
}
/*****************************************************************************
 * add a new authority record to a packet.
 *
 *  @param pkt     pointer to a dns packet structure.
 *  @param domian  domain for this RR
 *  @param type    type for this RR, currently T_A and T_PTR are supported.
 *  @param class   class for this RR, always set this to C_IN
 *  @param ttl     time-to-live for this RR
 *  @param datalen length of the area data points to.
 *  @param data    pointer to the data for this RR.
 *
 *  @return length of this RR or -1 on error.
 ****************************************************************************/
int dns_add_ar( dns_t * pkt, dns_rr * rr )
{
    int result = 0;

    result = dns_add_rr(pkt,rr);
    if( result < 0 ) return result;

    pkt->u.hdr.ancount = htons( ntohs(pkt->u.hdr.ancount) - 1);
    pkt->u.hdr.nscount = htons( ntohs(pkt->u.hdr.nscount) + 1);

    return result;

}

/*****************************************************************************
 * decode a dns query.
 *
 * The result from this function is valid until the next call to it.
 *
 * We do not (yet) decode the answer section in the question as required
 * for correct operation.
 * 
 * @param query      pointer to the dns query
 * @param querysize  length of the query message.
 *
 * @return a structure describing the query, NULL on error.
 ****************************************************************************/
dns_query * dns_decode_query(u_char * query , int querysize) {

  static dns_query result;
  int count = 0;
  u_char * data = NULL;

  HEADER * hdr = (HEADER *)query;

  /* initialize the result structure */
  memset( &result , 0, sizeof(dns_query));
  result.id = hdr->id;
  result.want_recursion = hdr->rd;

  if( querysize < sizeof(HEADER) ) {
	debug("dns_parse_query() : invalid size of %d bytes\n" , querysize);
	return NULL;
  }
 
  /* is this a query ? */
  if(hdr->qr != 0) {
    debug("dns_parse_query() : message is not a query\n" );
    return NULL;
  }

  if( (hdr->opcode != QUERY ) && (hdr->opcode != IQUERY )) {
    debug("dns_parse_query() : message is neither QUERY nor IQUERY\n" );
    return NULL;
  }

  data = query + sizeof(HEADER);

  if( hdr->opcode == QUERY ) {
    int qc = 0;
    int n  = 0;
    
    if( ntohs(hdr->qdcount) > MAXQUEST_PER_QUERY ) {
      qc = MAXQUEST_PER_QUERY;
      debug("dns_parse_query() : to many questions, truncating\n");
    } else {
      qc = ntohs(hdr->qdcount);
    }

    result.nquest = qc;

    for( n = 0; n < qc; n ++ ) { 

      count = dn_expand( query, query+querysize,
		         data,
		         result.question[n].query_arg , 
			 sizeof(result.question[n].query_arg));
      if( count < 0 ) {
	  return NULL;
      }
      data += count;
      GETSHORT(result.question[n].query_type, data );
      GETSHORT(result.question[n].query_class, data );
    }

    return &result;
  }

  /* we do not handle inverse queries. */
  debug("dns_parse_query() : won't handle inverse query\n");
  return NULL;

}

/*****************************************************************************
 * 
 * Walk over a dns message invoke a callback foreach RR found in the
 * answer section of the packet.
 *
 *****************************************************************************/
void dns_walk_packet( struct udp_packet *udp_pkt ,
		     dnswalk_cb callback ,
		     void * userdata )
{
     dns_walk_buf(udp_pkt->buf,udp_pkt->len,callback,userdata);
}

void dns_walk_buf  ( u_char * buf, int len ,
		     dnswalk_cb callback ,
		     void * userdata )
{
    u_char * data      = buf;
    u_char * query     = buf;
    int      querysize = len;
    HEADER * hdr       = (HEADER *)data;

    u_short  dtype;
    u_short  dclass;
    u_long   dttl;
    char     dname[MAXDNAME];
    
    data = query + sizeof(HEADER);

    if( hdr->opcode == QUERY ) {
      int qc = 0;
      int n  = 0;
      int count = 0;
    
      qc = ntohs(hdr->qdcount);

      debug("got %d queries\n",qc);

      for( n = 0; n < qc; n ++ ) { 

        count = dn_expand( query, query+querysize,
		           data,
		           dname , 
			   MAXDNAME);
        if( count < 0 ) {
  	    return ;
        }
        data += count;
        GETSHORT(dtype, data );
        GETSHORT(dclass, data );

	callback(IN_QUERY_SECT, dname, dtype, dclass, 0 , 0, NULL, userdata);
      }
    }

    /**
     * process answers
     */
    {
        int ac    = 0;
	int n     = 0;
	int count = 0;
	int len   = 0;
	u_char *  rr_data;

	ac = ntohs(hdr->ancount);

	debug("got %d answers\n",ac);

        for( n = 0; n < ac; n ++ ) { 
            count = dn_expand( query, query+querysize,
		               data,
		               dname , 
			       MAXDNAME);
            if( count < 0 ) {
  	        return ;
            }
            data += count;
 
            GETSHORT(dtype, data );
            GETSHORT(dclass, data );
            GETLONG (dttl  , data );
	    GETSHORT(len   , data );
	    rr_data = data;
	    data += len;

	    callback(IN_ANSWER_SECT, dname , dtype, dclass, dttl , len, rr_data , userdata);
	}
    }
    
    return ;
}

int dump_cb ( dns_sect section,
	      char * domain,
	      int type, int class, long ttl ,
	      int rr_len ,u_char * rr_data, 
	      void * user_data ) 
{
    debug( "%d|%d|%s\n" , type , class, domain );
    return 0;
}

