/*
  **
  ** dns.h - data structures used in dns queries and answers.
  **
  ** Part of the drpoxy package by Matthew Pratt. 
  **
  ** Copyright 1999 Matthew Pratt <mattpratt@yahoo.com>
  **
  ** This software is licensed under the terms of the GNU General 
  ** Public License (GPL). Please see the file COPYING for details.
  ** 
  **
*/

#ifndef DNS_H
#define DNS_H

#include <config.h>

#include <stdio.h>

#include <netinet/in.h>
#include <arpa/nameser.h>
#include <arpa/inet.h>
#include <resolv.h>

#include "tmdns.h"

#define MAXDOMAINS 128
#define DNSDATASIZE 4096

#define MAXQUEST_PER_QUERY	32

struct udp_packet {
  int len;
  struct in_addr   src_ip;
  int              src_port;

  socklen_t        src_len;
  struct sockaddr  src_address;

  int from_local;
  int from_mcast;
  u_char buf[4096];
};



typedef struct {
	short query_type;      		/* query type, usually T_A or T_PTR */
	short query_class;     		/* query_class, alway C_IN          */
	char query_arg[MAXDNAME];	/* object to search.                */
} dns_question;

typedef struct {
	short id;			  /* the query id for this query      */
	int want_recursion;		  /* 1 if the client wants recursion  */
	int inverse;	    		  /* 1 if invers query, 0 else        */
	int nquest;                       /* number of wuestions in this query*/
	dns_question question[MAXQUEST_PER_QUERY]; /* Questions within this query      */
} dns_query;


typedef struct {
  struct in_addr dst_ip;	       /* to wich address send this msg. */
  int           dst_port;	       /* to which port send             */

  socklen_t        dst_len;
  struct sockaddr  dst_address;

  int		from_mcast;            /* query came from multicast socket*/
  int           from_local;            /* query came from localhost       */
  int           fd;                    /* file descriptor on which we got */
  				       /* the query.                      */
  int           timeout;               /* time when to send back the answer*/

  int		bufsize;	       /* the size of the data buffer */
  u_char *	data;		       /* pointer to the first free byte in 
				          the buffer. */
  u_char *	last_dnptr;	       /* pointer to the last used entry in
				          dnptrs. */
  u_char *	dnptrs[MAXDOMAINS];    /* array of already compressed domains */
  union dns_msg {
      HEADER	hdr;		       /* the DNS header */
      u_char 	raw[DNSDATASIZE];      /* data buffer    */
  } u;
} dns_t;

typedef struct {
    char * name_server;		       /* name server for the domain. */
    char * admin;		       /* mailbox of the administrator */
    long serial;		       /* serial number for the data   */
    long refresh;		       /* refresh time for zone data   */
    long retry; 		       /* time to wait before a faild
					  refresh should be retried.   */
    long expire;		       /* time until zone data expires */
    long min_ttl;		       /* minitmum ttl value for all RR's*/
} dns_soa_rr;

typedef struct {
    short preference;			/* preference value for this mail host. */
    char * mail_host;			/* mail host. */
} dns_mx_rr;


typedef struct {
    char * cpu_type;			/* cpu type of this machine */
    char * os_name;			/* os of that machine       */
} dns_hinfo_rr;

#define DNSMAXTXT 32
typedef struct {
    int    nstrings;			/* number of strings */
    char * strings[DNSMAXTXT];		/* pointers to the strings */
} dns_txt_rr;

typedef struct {
    int priority;			/* priority for this svc */
    int weight;				/* weight for this service */
    int port;				/* port number */
    char * target;			/* target. No domain encoding in the binary rec*/
} dns_srv_rr;

typedef struct {
    char  * domain;
    short   type;
    short   class;
    long    ttl;
    void    (* freeFunc)();
    union {
	char *         dn;
	struct in_addr a;
	dns_soa_rr     soa;
	dns_mx_rr      mx;
	dns_hinfo_rr   hinfo;
	dns_txt_rr     txt;
	dns_srv_rr     srv;
    } rr ;
} dns_rr;


void dns_init( dns_t * );
dns_query * dns_decode_query(u_char * query , int querysize);

int  dns_init_answer( dns_t * , dns_query * );

int  dns_add_rr( dns_t * pkt , dns_rr * rr );

int  dns_add_qr( dns_t * pkt , char * domain,
		 short type, short class);

int dns_add_raw_rr( dns_t * pkt, char * domain,
		    short type , short class, long ttl ,
		    int rr_len, u_char * rr_data );

int  dns_get_len(dns_t *);

int udp_is_bridgesock( int fd );

typedef enum {
  IN_QUERY_SECT ,
  IN_ANSWER_SECT,
  IN_AUTHORITY_SECT,
  IN_EXTRA_SECT
} dns_sect;

typedef int (* dnswalk_cb)(dns_sect section, 
			   char * domain, 
			   int type, int class, long ttl,
			   int rr_len ,u_char * rr_data, 
			   void * user_data ) ;

extern dump_cb(dns_sect section, 
			   char * domain, 
			   int type, int class, long ttl,
			   int rr_len ,u_char * rr_data, 
			   void * user_data ) ;

void dns_walk_buf   ( u_char * buf , int len,
		      dnswalk_cb callback ,
		      void * userdata );

void dns_walk_packet( struct udp_packet *udp_pkt ,
		      dnswalk_cb callback ,
		      void * userdata );


#ifdef AF_INET6
#define HAVE_IP6
#endif

#endif
/* EOF */
