
#include <stdio.h>

#include <sys/utsname.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <ctype.h>

#include <assert.h>

#include "debug.h"
#include "tmdns.h"
#include "dns.h"
#include "llist.h"
#include "info.h"

#include "conf.h"

#include "ifaddrs.h"

static list_t * records;

static dns_rr * newHinfoRec(char * domain );
static dns_rr * newARec( char * domain , struct in_addr * ip );


/*****************************************************************************
 * Escape characters in domain labels.
 *
 * Returns a malloc'ed copy of the string with "." "\" and " " escaped.
 *
 *****************************************************************************/
char * escapeDomainLabel(char * label) {
    int needed = 0;
    char * now = NULL;
    char * to  = NULL;
    char * result = NULL;

    /* count how many characters we will need */
    for( now = label; *now != 0; now ++ ) {
	switch( *now ) {
	    case '.':
	    case '\\':
		needed += 2;
		break;
	    case ' ':
		needed += 4;
		break;
	    default:
		needed ++;
	}
    }

    result = (char *)malloc(needed + 1);
    to = result;

    for( now = label; *now != 0; now ++ ) {
	*to = 0;
	switch( *now ) {
	    case '.':
	    case '\\':
		*to = '\\';
		to ++;
		*to = *now;
		break;
	    case ' ':
		strcat(to,"\\032");
		to += 4;
		break;
	    default:
		*to = *now;
		to ++;
	}
    }

    * to = 0;

    return result;

}

/*****************************************************************************
 * Constructor and destructor functions for basic resource records.
 *
 *****************************************************************************/
void free_rr(dns_rr * rr) {
  free(rr);
}

void init_rr(dns_rr * rr,char * domain, short type ) {
  rr->domain   = strdup(domain);
  rr->type     = type;
  rr->class    = C_IN;
  rr->ttl      = config.default_ttl;
  rr->freeFunc = free_rr;
}

/*****************************************************************************
 * Constructor and destructor functions for hostinfo resource records.
 *
 *****************************************************************************/
void freeHinfoRec(dns_rr * hinfo) {
   free(hinfo->rr.hinfo.cpu_type);
   free(hinfo->rr.hinfo.os_name);
   free_rr(hinfo);
}

dns_rr * newHinfoRec(char * domain ) {

    struct utsname info;
    dns_rr * hinfo = NULL;

    char * os  = NULL;

    uname(&info);

    asprintf(&os, "%s %s" , info.sysname , info.release);

    hinfo = (dns_rr *)malloc( sizeof(dns_rr) );

    init_rr(hinfo,domain,T_HINFO);
    hinfo->rr.hinfo.cpu_type = strdup(info.machine);
    hinfo->rr.hinfo.os_name  = os;

    hinfo->freeFunc = freeHinfoRec;

    return hinfo;
}

/*****************************************************************************
 * Constructor function for ipv4 address resource records.
 *
 *****************************************************************************/
dns_rr * newARec( char * domain , struct in_addr * ip ) {
    dns_rr * aRec = NULL;
    
    aRec = (dns_rr *)malloc( sizeof(dns_rr) );
    init_rr(aRec,domain,T_A);
    memcpy(&(aRec->rr.a),ip,sizeof(struct in_addr));

    aRec->freeFunc = free_rr;

    return aRec;
}

/*****************************************************************************
 * Constructor and destructor functions for pointer resource records.
 *
 *****************************************************************************/
void freePtrRec(dns_rr * rr) {
    free(rr->rr.dn);
    free_rr(rr);
}

dns_rr * newPtrRec( char * domain , char * rev ) {
    dns_rr * ptrRec = NULL;
    
    ptrRec = (dns_rr *)malloc( sizeof(dns_rr) );
    init_rr(ptrRec,rev,T_PTR);
    ptrRec->rr.dn = strdup(domain);

    ptrRec->freeFunc = freePtrRec;

    return ptrRec;
}

/*****************************************************************************
 * Constructor and destructor functions for service resource records.
 *
 *****************************************************************************/
void freeSrvRec(dns_rr * rr) {
    free(rr->rr.srv.target);
    free_rr(rr);
}

dns_rr * newSrvRec( char * domain,
                  unsigned short priority,
                  unsigned short weight,
                  unsigned short port,
                  char * target )
{
    dns_rr * srvRec = NULL;

    srvRec = (dns_rr *)malloc( sizeof(dns_rr) );
    init_rr(srvRec,domain,T_SRV);

    srvRec->rr.srv.target   = strdup(target);
    srvRec->rr.srv.priority = priority;
    srvRec->rr.srv.weight   = weight;
    srvRec->rr.srv.port     = port;

    srvRec->freeFunc = freeSrvRec;

    return srvRec;
}

/*****************************************************************************
 * Read a service regestry and add a service record for each entry.
 *
 *****************************************************************************/
void info_read_serviceconf( char * service_file , char * hostname ) {

    FILE *fp;
    char line[1024];
    int  lineno = 0;

    debug("read service config %s ...\n", service_file);

    fp = fopen (service_file, "r");
    if (!fp) {
      debug_perror("no service file");
      return;
    }

    while (fgets(line, 1024 , fp)) {

    	char * now  = line ;
	int fields  = 0;

        char proto[1024];
	char service[1024];
        int port     = -1;
        int prio     = -1;
        int weight   = -1;
        char * name  = NULL;
	int name_ofs = 0;

	char * serv_name = NULL;
	char * ptr_name  = NULL;

	lineno ++;

        /* skip whitespace */
        while( isblank(*now) ) { now ++; }

        line[strlen(line) - 1] = 0; /* kill '\n' */

        if ( *now =='#') { continue; } /* skip lines with comment */
        if ( *now == 0 ) { continue; } /* skip empty lines */


	fields = sscanf(now,"%s %d %s %d %d %n", proto , &port, service, &prio, &weight , &name_ofs );
	
	if( ( fields < 5 ) && ( fields > 0 ) ) {
	    debug("syntax error in line %d\n",lineno);
	}

	if( name_ofs > 0 ) {
	    name = now + name_ofs;
	}

	debug("got %d fields, proto=%s port=%d service=%s prio=%d weight=%d name=%s (ofs=%d)\n",
		fields, proto , port, service, prio, weight, name, name_ofs );

	/* semantic check ... */

	if( *name != 0 ) {
	    char * escaped = NULL;
	    escaped = escapeDomainLabel(name);
	    if( strlen(escaped) > 63 ) {
		debug("truncate service name becaues it longer than 32 chars.\n");
		escaped[63] = 0;
	    }
	    asprintf(&serv_name, "%s._%s._%s.local",escaped,service,proto);
	    free(escaped);
	    asprintf(&ptr_name , "_%s._%s.local",service,proto);
	} else {
	    asprintf(&serv_name, "_%s._%s.local",service,proto);
	}

        ll_add(records, newSrvRec(serv_name, prio, weight, port, hostname ));
	if( ptr_name != NULL ) {
            ll_add(records, newPtrRec(serv_name,ptr_name));
	}

	free(serv_name);
	if( ptr_name != NULL ) { free(ptr_name); }

    }

    fclose(fp);
    return; 
}
/*****************************************************************************
 * initialize the resource database.
 *
 *  This function adds resource records we want to publish to the record 
 *  registry. It will perform the following steps :
 *
 *  - get the hostname from the config file or from utsname
 *  - get all interface addresses and add an A record for each address.
 *  - add a PTR rec for each address to allow reverse lookup.
 *  - add a HINFO record for hostname.local.
 *  - call a function to add service records to the registry.
 *
 *****************************************************************************/

void info_init() {

    struct ifconf ifconfig;
    struct ifaddrs * ifs = NULL;
    struct ifaddrs * ifnow = NULL;

    struct utsname info;
    char * namebuf = NULL;
    char * raw_namebuf = NULL;
    char * dot_pos = NULL;

    debug("Initialize host data ...\n");

    uname(&info);

    if( ( config.hostname != NULL ) && (  config.hostname[0] != 0 ) ) {
        raw_namebuf = strdup(config.hostname);
    } else {
        raw_namebuf = strdup(info.nodename);
    }

    /* only first label */
    if( (dot_pos = index(raw_namebuf,'.')) != NULL ) {
	*dot_pos = 0;
    }

    asprintf(&namebuf,"%s.local", raw_namebuf );
    free(raw_namebuf);
    
    //asprintf(&namebuf,"%s.local", "andinc");
    
    records = ll_new();

    /*
     * Find our ethernet interfaces and add A records for
     * each IPv4 address we have.
     * Also add PTR records for the reverse zone.
     */
    getifaddrs(&ifs);

    for(ifnow = ifs; ifnow; ifnow = ifnow->ifa_next) {
	if( ifnow->ifa_flags & IFF_LOOPBACK ) continue;

	debug("got address on interface %s\n", ifnow->ifa_name );

	switch( ifnow->ifa_addr->sa_family ) {
	  case AF_INET:
		{
		    struct in_addr inaddr;
		    unsigned char * ib = (unsigned char *)&inaddr;
		    char * revname = NULL;

		    struct sockaddr_in * addr = 
			    (struct sockaddr_in *)ifnow->ifa_addr;

		    memcpy(&inaddr, &(addr->sin_addr) , sizeof(struct in_addr));
		    debug("  %s\n", inet_ntoa(inaddr) );
    		    ll_add(records , newARec(namebuf,&inaddr));

		    asprintf( &revname , "%d.%d.%d.%d.in-addr.arpa" ,
				    ib[3] , ib[2], ib[1], ib[0] );
		    debug("Reverse addrees is %s\n" , revname );
		    ll_add(records , newPtrRec(namebuf,revname));
		    free(revname);
		}
		break;

	  case AF_INET6:
		/* TODO */
		break;
	}
    }


    freeifaddrs(ifs);

    //ll_add(records, newSrvRec("_http._tcp.aule.local" , 0 , 1 , 80 , "aule.local."));
    ll_add(records , newHinfoRec(namebuf));

    if( config.service_file[0] != 0 ) {
        info_read_serviceconf(config.service_file, namebuf );
    }
    if( config.dynamic_service_file[0] != 0 ) {
        info_read_serviceconf(config.dynamic_service_file, namebuf );
    }

    free(namebuf);
}

/*****************************************************************************
 * Destroy our RR registry.
 *
 * 
 *****************************************************************************/
void info_destroy() {
    ll_entry_t * el      = NULL;

    for( el = ll_first(records) ; el != NULL ; el = ll_next(el) ) {
	dns_rr * rr = (dns_rr *)el->data;
	assert(rr != NULL );
	if( rr->freeFunc != NULL ) {
	    rr->freeFunc(rr);
	} else {
	    free_rr(rr);
	}
    }

    ll_delete(records);
    records = NULL;
}

/*****************************************************************************
 * Init a search state structure.
 *
 *
 *****************************************************************************/
void info_init_search( search_state * state, char * query, int type ) {
    
    assert(state != NULL);

    state->dname = query;
    state->type  = type;
    state->data  = NULL;
    state->el    = NULL;

}

/*****************************************************************************
 * search for an RR.
 *
 *   The search_state structure should must be filled out by the 
 *   info_init_search() function before calling this function. The 'el' member
 *   in the search stucture is updated to allow enumeration of all matching
 *   records.
 *   
 * Arguments:
 *   state  - state of this search.
 *
 * Note:
 *   A special case is state->dname == NULL. This is used to find all 
 *   resource records we want to be authorative for. These are the  A and AAAA 
 *   records for now.
 *
 * Result:
 *   1 if any record has been found, 0 otherwise. The search structure is 
 *   updated to reflect the position in the list where the record has been 
 *   found.
 *
 *****************************************************************************/
int info_search( search_state * state ) {

    char * dname;
    short type;
    ll_entry_t * startAt = NULL;
    ll_entry_t * el      = NULL;

    dname   = state->dname;
    type    = state->type;

    if( state->el == NULL ) {
	startAt = ll_first(records);
    } else {
	startAt = ll_next(state->el);
	if( startAt == NULL ) {
	    state->el = NULL;
	    state->data = NULL;
	    return 0;
	}
    }
    
    if( state->el != NULL );

    if( records == NULL ) {
	debug("Records array not initialized\n");
        return 0;
    }

    for( el = startAt ; el != NULL ; el = ll_next(el) ) {
	dns_rr * rr = el->data;

	/*debug("Search %s, have %s type %d\n", state->dname, rr->domain, rr->type ); */

	if( ( (state->type == T_ANY) || (rr->type == state->type) ) && 
	    ( (state->dname == NULL) || (strcmp(state->dname,rr->domain) == 0) ) ) 
	{
	
	    if( (state->dname == NULL) && !(( rr->type == T_A)||(rr->type == T_AAAA))) {
		continue;
	    }
	    
	    state->data = rr;
	    state->el   = el;
	    return 1;
	}
    }

    return 0;

}

/***
 * this callback should be called when we are probing for our RR's
 * and we got an answer for any of our queries.
 *
 * We will then remove the offending RR's before we send another probe
 */
int info_conflict_cb ( 
	      dns_sect section,
              char * domain,
              int type, int class, long ttl,
              int rr_len ,u_char * rr_data,
              void * user_data )
{   

    search_state s;

    if( (section == IN_QUERY_SECT) || (section == IN_EXTRA_SECT) ) {
	return 0;
    }
    debug( "conflict for %d|%d|%s ?\n" , type , class, domain );

    s.dname = domain;
    s.type  = type;
    s.el    = NULL;
    s.data  = NULL;

    while( info_search(&s) >0 ) {
	debug("got conflict\n");
	syslog(LOG_ERR, "conflict on name \"%s\" , type %d - drop my record\n",domain,type);
	((dns_rr *)s.el->data)->freeFunc(s.el->data);
	ll_remove(s.el);
	s.el = NULL;
    }
    
    return 0;
}

