
/*
  **
  ** serv_udp.c
  **
  ** Part of the dproxy package by Matthew Pratt. 
  **
  ** Copyright 1999 Matthew Pratt <mattpratt@yahoo.com>
  **
  ** This software is licensed under the terms of the GNU General 
  ** Public License (GPL). Please see the file COPYING for details.
  ** 
  **
*/

/*
  **
  ** See ChangeLog for version history.
  **
*/

#include <config.h>

#include <net/if.h>

#include "tmdns.h"
#include "serv_udp.h"
#include "debug.h"
#include "ifaddrs.h"
#include "conf.h"

#define MCAST_V4_ADDR "224.0.0.251"
#define UNICAST_LOCAL "127.0.0.1"

static int mcast_sock = -1;
static int local_sock = -1;
static int bridge_sock = -1;

static struct sockaddr mcast_v4_sa;

/*****************************************************************************
 * open sockets we ar interested in.
 *
 * We will open
 *  - one unicast listener on port 53, loopback device (if enabled)
 *  - one unicast listener per interface that is not loopback
 *  - one multicast listener for the IPv4 mDNS group.
 *
 * for IPv6 we should also open a socket for the IPv6 multicast group
 *
 * Arguments:
 *   exclude_devs  :  device names to exclude when geting the addresses.
 *   		      not used yet.
 *   sockets       :  Array of intergers where to store the file
 *   		      desctiptors.
 *
 * Returns:
 *   number of fd's opened.
 *
 ****************************************************************************/
int udp_open_sockets(char ** exclude_devs , int * sockets) {

    int   iffound = 0;

    struct ifaddrs * ifs = NULL;
    struct ifaddrs * ifnow = NULL;


    int idx;
    int sockidx = 0;

    if( config.dns_bridge ) {
	struct sockaddr_in * sock = (struct sockaddr_in *)&mcast_v4_sa;
	struct in_addr mcast_addr;

	memset( sock , 0 , sizeof(struct sockaddr_in) );
	sock->sin_family = AF_INET;
	sock->sin_port = htons(config.dns_port);

	inet_aton( UNICAST_LOCAL , &(sock->sin_addr) );
	sockets[sockidx] = udp_sock_open( 0 , (struct sockaddr *)sock);

	if( sockets[sockidx] >= 0 ) {
	    bridge_sock = sockets[sockidx];
	    debug("bridge socket is index %d, fd = %d\n", sockidx , bridge_sock );
	    sockidx ++;
	} else {
	    debug_perror("can not create local dns listener socket");
	    syslog(LOG_ERR,"can not create local dns listener socket for %s", 
			    udp_sockaddr_str((struct sockaddr *)sock));
	}
 
    }

   /**
     * open a socket for each interface address.
     *
     * We need this because we can not simply listen to ANY_ADDR
     *
     * We need to distinguish between mcast and unicast
     */
    getifaddrs(&ifs);

    for(ifnow = ifs; ifnow; ifnow = ifnow->ifa_next) {
	switch( ifnow->ifa_addr->sa_family ) {
	    case AF_INET:
		{
		    struct sockaddr_in * addr = (struct sockaddr_in *)ifnow->ifa_addr;

		    struct sockaddr      sc;
		    struct sockaddr_in * sa = (struct sockaddr_in *)&sc;    

  		    sa->sin_family = AF_INET;
		    sa->sin_port = htons(config.port);
		    memcpy((void *)&(sa->sin_addr), 
			   (void *)&(addr->sin_addr), sizeof(struct in_addr));

		    sockets[sockidx] = udp_sock_open( 0 , &sc );
		    if( sockets[sockidx] >= 0 ) {
			if( ifnow->ifa_flags & IFF_LOOPBACK ) {
			  local_sock = sockets[sockidx];
			}
			sockidx ++;
		    } else {
			int err = errno;
			debug("can not open listener socket for %s : %s\n" , 
				udp_sockaddr_str(&sc), strerror(err) );
			syslog(LOG_ERR, "can not open listener socket for %s : %s\n" , 
				udp_sockaddr_str(&sc), strerror(err) );
		    }
		}
		break;
	
	    case AF_INET6:
		break;
	}
    }

    /**
     * open the ipv4 multicast socket.
     *
     * Seems that this should be the last socket we open. Once this was the
     * first thing I did, but then I got packets on the bridge socket which
     * not even have the same port number ! Kernel bug ?
     */
    {
	struct sockaddr_in * sock = (struct sockaddr_in *)&mcast_v4_sa;
	struct in_addr mcast_addr;

	memset( sock , 0 , sizeof(struct sockaddr_in) );
	sock->sin_family = AF_INET;
	sock->sin_port = htons(config.port);

	inet_aton( MCAST_V4_ADDR , &(sock->sin_addr) );
	sockets[sockidx] = udp_sock_open( 1 , (struct sockaddr *)sock);
	
	if( sockets[sockidx] >= 0 ) {
	    mcast_sock = sockets[sockidx];
	    debug("mcast socket is index %d, fd = %d\n", sockidx , mcast_sock );
	    sockidx ++;
	} else {
	    int err = errno;
	    debug("can not crate multicast listener socket for %s : %s\n" , 
		  udp_sockaddr_str((struct sockaddr *)sock), strerror(err) );
	    syslog(LOG_ERR, "can not open listener socket for %s : %s\n" , 
		  udp_sockaddr_str((struct sockaddr *)sock), strerror(err) );
	}
 
    }

 
    return sockidx ;
}
/*****************************************************************************
 * Actually open a single socket.
 *
 * Arguments are
 *  mcast   -  !=0 when this should be a multicast socket, else 0
 *  sock    -  sockaddr structure giving the address to listen on
 * 
 * Returns
 *   >= if the socket could be opened, < 1 if an error occured.
 *
 ****************************************************************************/
int udp_sock_open( int mcast , struct sockaddr * sock )
{
  int fd;
  
  fd = socket( sock->sa_family , SOCK_DGRAM, IPPROTO_UDP);

  /* Error */
  if( fd < 0 ){
	 debug_perror("Could not create socket");
	 syslog(LOG_ERR,"Could not create socket : %s\n",
			 strerror(errno));
	 return -1;
  } 

  /* bind() the socket to the interface */
  if (bind(fd, sock , sizeof(struct sockaddr)) < 0){
	 debug_perror("Could not bind to port");
	 syslog(LOG_ERR,"Could not create socket : %s\n",
			 strerror(errno));
	 close(fd);
	 return -1;
  }

  if( mcast ) {
    if( sock->sa_family == AF_INET ) {
      struct sockaddr_in * sa = (struct sockaddr_in *) sock;
      struct ip_mreq mreq;
      unsigned char ttl  = 255;
      unsigned char loop = 0;
      memcpy(&mreq.imr_multiaddr,
             &(sa->sin_addr),
             sizeof(struct in_addr));
      mreq.imr_interface.s_addr = INADDR_ANY;

      /* Do not loop back the query to myself. */
      setsockopt(fd, IPPROTO_IP, IP_MULTICAST_LOOP, &loop, sizeof(loop));

      /* spec require a ttl of 255. */
      setsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL , &ttl , sizeof(ttl));

      /* join the group */
      setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq));
    }
  }
             
  return(fd);
}

/*****************************************************************************/
int udp_packet_read(int sockfd, struct udp_packet *udp_pkt )
{
  int numread;
  struct sockaddr_in * sa;

  /* Read in the actual packet */
  udp_pkt->src_len = sizeof(struct sockaddr);

  if ((numread = recvfrom(sockfd, &udp_pkt->buf, sizeof(udp_pkt->buf),0,
	&(udp_pkt->src_address), &(udp_pkt->src_len) )) < 0) {
    debug_perror("udp_packet_read: recvfrom");

    return -1;
  }

  /* Then record where the packet came from */
  /* WILL GO AWAY */
  sa = (struct sockaddr_in *)&(udp_pkt->src_address);
  memcpy((void *)&udp_pkt->src_ip, (void *)&(sa->sin_addr), sizeof(struct in_addr));
  udp_pkt->src_port = ntohs(sa->sin_port);

  if( sockfd == local_sock ) { udp_pkt->from_local = 1; };
  if( sockfd == mcast_sock ) { udp_pkt->from_mcast = 1; };

  udp_pkt->len = numread;

  return numread;
}
/*****************************************************************************/
int udp_send_dnsmsg( int sockfd, dns_t * pkt ) {

  struct sockaddr_in sa;

  /*
   * First reply to the multicast address. 
   */
  if( pkt->from_mcast || pkt->dst_port < 0 ) {
      sendto( mcast_sock, 
	      &pkt->u.raw , dns_get_len(pkt), 
	      0 , 
	      &mcast_v4_sa , sizeof(mcast_v4_sa));

      debug("sent message of %d bytes to %s\n" , 
	    dns_get_len(pkt) , udp_sockaddr_str(&mcast_v4_sa) );
  }


  /*
   * draft requires that we send an unicast packet back to the
   * sender if its source port is not 5353. Maybe we should also
   * look if it was addressed to the multicast address/socket ?
   */
  if( (pkt->dst_port >= 0) && (pkt->dst_port != config.port ) ) {
      int bytes = 0;
      /* Zero socket address */
      memset((void *)&sa, 0, sizeof(sa));

      bytes = sendto(sockfd, &pkt->u.raw , dns_get_len(pkt), 
	      0, &(pkt->dst_address), pkt->dst_len );

      debug("sent %d(of %d) bytes to %s\n" , 
	    bytes , dns_get_len(pkt) , udp_answerdst2str(pkt) );
  }

}
/*****************************************************************************
 * return true if the given socket is the/a dns bridge socket.
 *
 *****************************************************************************/
int udp_is_bridgesock( int fd ) {
    if( fd == bridge_sock ) {
	return 1;
    }
    return 0;
}

/*****************************************************************************
 * Copy address from an incomming udp_packet to an answer packet.
 *
 *
 *****************************************************************************/
void udp_copy_answer_address(dns_t * answer, struct udp_packet * udp_pkt) {

    memcpy( (void *)&(answer->dst_address),
            (void *)&(udp_pkt->src_address),
                      udp_pkt->src_len);
    answer->dst_len  = udp_pkt->src_len;

    /* should go away ... */
    answer->dst_port = udp_pkt->src_port;
  
}

/*****************************************************************************
 * Take source address and port from a struct sockaddr and return a
 * string suitable for logging/debugging.
 *
 * The result of this function is a pointer to a static buffer that will be
 * re-used in the next call to this function.
 *
 *****************************************************************************/
#ifndef INET6_ADDRSTRLEN
#define INET6_ADDRSTRLEN 256
#endif
#define PORT_LEN 7
char * udp_sockaddr_str( struct sockaddr * addr ) {

    static char result[INET6_ADDRSTRLEN+PORT_LEN+2];
    int  addr_port;
    char port_s[PORT_LEN];

    switch( addr->sa_family ) {
#ifdef HAVE_INET_NTOP
        case AF_INET:
	case AF_INET6:
	    inet_ntop(addr->sa_family, &(((struct sockaddr_in *)addr)->sin_addr),result,INET6_ADDRSTRLEN);
#else
        case AF_INET:
	    strncpy(result,inet_ntoa((struct sockaddr_in *)addr),INET6_ADDRSTRLEN);
#endif
	    addr_port = htons(((struct sockaddr_in *)addr)->sin_port);
	    snprintf(port_s,PORT_LEN,"#%d",addr_port);
	    strncat(result,port_s,INET6_ADDRSTRLEN+PORT_LEN+2);
	    return result;
	    break;
	default:
	    return "(unknown address family)";
    }

}

char * udp_answerdst2str(dns_t * answer) {
    return udp_sockaddr_str( &(answer->dst_address) );
}

char * udp_pktsrc2str(struct udp_packet * udp_pkt) {
    return udp_sockaddr_str( &(udp_pkt->src_address) );
}

