
#include <config.h>

#include "tmdns.h"
#include "conf.h"
#include "serv_udp.h"
#include "debug.h"
#include "dns.h"
#include "info.h"

#include "llist.h"

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#include <pwd.h>

#include <sys/time.h>
#include <sys/types.h>

#include <assert.h>

/*****************************************************************************
 * Global variables and prototypes.
 *
 *****************************************************************************/

int  handle_query(int sock, struct udp_packet * udp_pkt);
int  store_query(int sock, struct udp_packet * udp_pkt);
int  answer_outstanding_queries(); 
void fill_outstanding_queries(int sock, struct udp_packet * udp_pkt); 

int go_down = 0;
int do_config = 0;

/* a list of domain name suffixes that may be queried per multicast */
char * dot_local_domains[] = {
    ".local" ,
    ".254.169.in-addr.arpa" ,
    "0.8.e.f.ip6.arpa" ,
    NULL
};

/* list of queries to answer. */
list_t * query_list = NULL;

/*****************************************************************************
 * Signal handlers
 *
 *****************************************************************************/
void sig_hup (int signo)
{
  signal(SIGHUP, sig_hup); /* set this for the next sighup */
  /* conf_load (config.config_file); */
  do_config = 1;
}

void sig_term (int signo) {
  debug("signal %d, going down ...", signo);
  go_down = 1;
}

/*****************************************************************************
 * print usage informations to stderr.
 * 
 *****************************************************************************/
void usage(char * program , char * message ) {
  fprintf(stderr,"%s\n" , message );
  fprintf(stderr,"%s ver. %s\n" , PACKAGE , VERSION );
  fprintf(stderr,"usage : %s [-c <config-file>] [-bdFhP] [-p <port>\n",program );
  fprintf(stderr,"\t-b \tdisable dns bridge mode for local queries\n");
  fprintf(stderr,"\t-c <config-file>\tread configuration from <config-file>\n");
  fprintf(stderr,"\t-d \t\tturn on debug (=non-daemon) mode.\n");
  fprintf(stderr,"\t-F \t\tdo not fork, impiles -d. (for debuging only)\n");
  fprintf(stderr,"\t-p <port>\tlisten on port <port>.\n");
  fprintf(stderr,"\t-P \t\tprint configuration on stdout and exit.\n");
  fprintf(stderr,"\t-h \t\tthis message.\n");
}

/*****************************************************************************
 * get commandline options.
 * 
 * @return 0 on success, -1 on error, -2 if the progam should be terminated
 * with return code 0.
 *
 *****************************************************************************/
int get_options( int argc, char ** argv ) {

  char c = 0;
  int dns_bridge = 0;
  int not_daemon = 0;
  int no_fork    = 0;
  int want_printout = 0;
  int want_version  = 0;
  int debug_port = 0;
  char * progname = argv[0];
  char * confname = NULL;

  conf_defaults();

  while( (c = getopt( argc, argv, "bc:dFhp:Pv")) != EOF ) {
    switch(c) {
	case 'b':
		dns_bridge = 1;
		break;
	case 'c':
		confname = optarg;
		break;
	case 'd':
		not_daemon = 1;
		break;
	case 'F':
		not_daemon = 1;
		no_fork    = 1;
		break;
	case 'h':
		usage(progname,"");
		return -2;
	case 'p':
		debug_port = atoi(optarg);
		break;
	case 'P':
		want_printout = 1;
		break;
	case 'v':
		want_version = 1;
		break;
	default:
		usage(progname,"");
		return -1;
    }
  }

  if( confname == NULL ) confname = CONFIG_FILE_DEFAULT;
  conf_load(confname);

  if( dns_bridge ) {
	config.dns_bridge = 0 ;
  }
  
  /** unset daemon-mode if -d was given. */
  if( not_daemon ) {
	config.daemon_mode = 0;
  } else {
	config.daemon_mode = 1;
  }

  /** set no_fork if -F was given. */
  if( no_fork ) {
	config.no_fork = 1;
  }

  if( debug_port > 0 ) {
	config.port = debug_port;
  } else {
	config.port = PORT;
  }

  if( want_printout ) {
	conf_print();
	return -2;
  }

  if( want_version ) {
	fprintf(stderr,"this is tmdns %s\n", VERSION );
#ifdef DEBUG
	fprintf(stderr,"    debug support is enabled\n");
	fprintf(stderr,"    default debug log is %s\n", DEBUG_FILE_DEFAULT );
#else
	fprintf(stderr,"    debug support is disabled\n");
#endif
	fprintf(stderr,"    default config file is %s\n", CONFIG_FILE_DEFAULT );
	return -2;
  };

  return 0;

}

/*****************************************************************************
 * change userid.
 *
 * Beside setting the uid of thos process, we also change the owner of 
 * log- and pidfiles to allow writing when we are not longer root.
 * 
 *****************************************************************************/
void change_id() {

    struct passwd * pw = NULL;

    if( (config.username == NULL ) || (config.username[0] == 0 )) {
	debug("no username given\n");
	return;
    }

    pw = getpwnam( config.username );
    if( pw == NULL ) {
	debug("no user named %s\n", config.username );
	return;
    }

    /* change ownership of the pid file so we can unlink it later */
    if( (config.pid_file ) != NULL && (config.pid_file[0] != 0) ) {
	debug("set ownership of the pid file %s", config.pid_file );
	chown( config.debug_file , pw->pw_uid, pw->pw_gid );
    }

    /* change ownership of the log file */
    if( (config.debug_file ) != NULL && (config.debug_file[0] != 0) ) {
	debug("set ownership of the debug file %s", config.debug_file );
	chown( config.debug_file , pw->pw_uid, pw->pw_gid );
    }

    debug("drop privilegs: new uid = %d, new gid = %d\n",
	  pw->pw_uid, pw->pw_gid);

    setgid(pw->pw_gid);
    setuid(pw->pw_uid);
    return;
}

/*****************************************************************************/
int main(int argc, char **argv) {

  /* sockets to listen on. */
  int sockfds[ MAX_IF + 1 ];
  int usedifs = 0;

  /* count this down to check when we have send all probes we should have */
  int probing    = 3;	

  /* when to answer the next outstanding answer */
  int next_outstanding_answer = 0;

  int i = 0;

  char * listen_on[] = { NULL }; 

  fd_set readfds ;

  int numread;
  struct udp_packet pkt;
  int optres = 0;

  memset( &pkt , 0, sizeof(pkt) );

  /* get commandline options, load config if needed. */
  if( (optres = get_options( argc, argv )) < 0 ) {
	if(optres == -2) exit(0);
  	exit(1);
  }

  openlog("tmdns", LOG_PID , LOG_DAEMON );

  /* initisalize the RRs to announce */
  info_init();

  for( i = 0; i < MAX_IF ; i ++ )
    sockfds[i] = -1;

  debug("open sockets ...\n");
  usedifs = udp_open_sockets( listen_on , sockfds );

  if( usedifs < 1 ) {
	syslog(LOG_ERR, "no sockets to listen on.");
	debug("Sorry, no sockets to listen on.\n" );
	exit(2);
  }

  debug("have %d sockets\n", usedifs );

  /* initialize query backlog */
  query_list = ll_new();

  if (config.daemon_mode) {
    /* Standard fork and background code */
    switch (fork()) {
	 case -1:	/* Oh shit, something went wrong */
		debug_perror("fork");
		exit(-1);
	 case 0:	/* Child: close off stdout, stdin and stderr */
		close(0);
		close(1);
		close(2);
		setpgrp();
		break;
	 default:	/* Parent: Just exit */
		exit(0);
    }
  }

  /* create pid file */
  {
    FILE * pidfp = NULL;
    pidfp = fopen( config.pid_file , "w" );
    if( pidfp != NULL ) {
	fprintf(pidfp,"%d\n", getpid());
	fclose(pidfp);
    } else {
	syslog(LOG_ERR,"can't open pid file %s : %s\n",
		config.pid_file , debug_errmsg(errno) );
	debug("can not open pid file %s : %s\n",
		config.pid_file , debug_errmsg(errno) );
    }
  }

  /* drop su privilegs */
  change_id();

  signal(SIGHUP, sig_hup);
  signal(SIGTERM, sig_term);
  signal(SIGQUIT, sig_term);
  signal(SIGINT, sig_term);

  syslog(LOG_INFO,"enter main loop");


  while(go_down != 1){
    int numready = 0;
    int maxfd = 0;
    struct timeval tv;

     /* 
      * when the child process that handles the actual response exits, 
      * we must call wait so that we dont get a zombie 
      */

    FD_ZERO(&readfds);

    for( i = 0; i < usedifs; i ++ ) {
	if( sockfds[i] < 0 ) continue;
     	FD_SET( sockfds[i] , &readfds);
	if( maxfd < sockfds[i] ) maxfd = sockfds[i] ;
    }

    if( probing >= 0 ) {
      tv.tv_usec = 250000;
      tv.tv_sec  = 0;
    } else {
      tv.tv_usec = 0;
      tv.tv_sec  = 1;
    }

    //debug("select , timeout in %d.%6.6d sec.\n", tv.tv_sec, tv.tv_usec );
    numready = select( maxfd+1 , &readfds , NULL, NULL , &tv );

    if( numready <= 0 ) {
	
	if( do_config ) {
	    /*
	     * do_config would be set from the signal handler.
	     */
  	    conf_load (config.config_file);
	    info_destroy();
	    info_init();
	    do_config = 0;
	    probing = 3;
	    config.daemon_mode = 0;
	}

	if( numready != 0 ) {
	    if( errno != EINTR )
	    	debug_perror("error on select !");
	    continue;
	}

	if( probing == 0 ) {
	    /*
	     * this is the first timeout after we have send our last
	     * probe. Set probing to -1 to start normal
	     * operation.
	     */
	    announce(sockfds[0]);
	    probing = -1;
	}

	if( probing > 0 ) {
	    /*
	     * still in probing mode - send another probe 
	     */
	    probe(sockfds[0]);
	    probing --;
	}

	if( ll_first(query_list) != NULL ) {
	    /*
	     * Check outstanding queries that may should be answered now.
	     */
	     debug("handle outstanding queries ...\n");
	     next_outstanding_answer = answer_outstanding_queries();
	}

    } else {
	/* got some data ... */
 	for( i = 0; i < usedifs; i ++ ) {
  
	    int sockfd = -1;

	    debug("check socket idx %d, fd=%d\n", i , sockfds[i] );

	    if( sockfds[i] < 0 ) continue;
	    /* check if there is something to read */
	    if( ! FD_ISSET( sockfds[i] , &readfds) ) 
		continue;
	
	    sockfd = sockfds[i] ;

	    /* get the packet from the socket. */
  	    memset( &pkt , 0, sizeof(pkt) );
	    numread = udp_packet_read( sockfd, &pkt );	 
	    if( numread < 0 ) {
	        debug("no data ...\n");
	        continue;
	    }


	    debug("Message from %s , size %d bytes\n" , 
		      udp_pktsrc2str(&pkt), numread  );
	    debug("  from local = %d from mcast = %d\n" , 
		      pkt.from_local , pkt.from_mcast );
	
	    if(numread < sizeof(HEADER)+1 ) {
	        debug("invalid size : %d < %d\n",numread,sizeof(HEADER)+1);
		syslog(LOG_NOTICE,"dns packet from %s has invalid size of %d bytes.\n",
			udp_pktsrc2str(&pkt), numread  );
	        continue;
	    }


	    if(  probing < 0 ) {

		/* not probing */

		if( ((HEADER *)&(pkt.buf))->qr == 0 ) {

		    /* packet is a query */

		    if( udp_is_bridgesock(sockfd)) {
			debug("query from bridge socket\n");
  	      	        next_outstanding_answer = store_query(sockfd, &pkt );
		    } else {
  	      	        handle_query(sockfd, &pkt );
		    }
	        } else {

		    /* packet is an answer */

		    if( ll_first(query_list) != NULL ) {
			/*
			 * We have some outstanding queries, check if 
			 * we need to add rr's from this anser to
			 * the query.
			 */
	     		fill_outstanding_queries(sockfd,&pkt);

		    } else {
		        debug("ignore answer\n");
		    }
		}
	    } else {

		/* probing for dups */

		if( ((HEADER *)&(pkt.buf))->qr == 1 ) 
		{
		    /* got an answer while probing*/

		    debug("got answer for my query while probing for dups\n");

		    dns_walk_packet( &pkt , info_conflict_cb , NULL );
		}

	    }

        } /* for each ready fd */

	if( (next_outstanding_answer < time(NULL)) && ( ll_first(query_list) != NULL) ) {
	    /*
	     * Check outstanding queries that may should be answered now.
	     */
	     debug("handle outstanding queries ...\n");
	     next_outstanding_answer = answer_outstanding_queries();
	}


     } /* no error */
  } /* while not go down */
  unlink( config.pid_file );
  syslog(LOG_INFO,"normal exit");
}
/*****************************************************************************
 * Sleep for a random delay between 20-120 msec.
 *****************************************************************************/
void random_sleep() {
    unsigned long rdelay = 0;
    rdelay = 20000 + (100000.0*rand()/(RAND_MAX+1.0));
    debug("random sleep for %d usec\n" , rdelay );
    usleep(rdelay);
}
/*****************************************************************************
 * handle a query.
 *
 * - decode the query
 * - send the answer.
 * 
 * @param sock     socket to send back the answer.
 * @param udp_pkt  data packet that contains the query.
 *
 * @return >=0 on success, < 0 on error.
 *****************************************************************************/
int handle_query( int sock,  struct udp_packet *udp_pkt ) {

    int n = 0;
    int answer_count  = 0;
    dns_query * query = NULL;
    dns_t answer;
    
    dns_init(&answer);

    debug_dns("query from client", &(udp_pkt->buf));

    udp_copy_answer_address(&answer,udp_pkt);

    debug("decode query ...\n");
    query = dns_decode_query(udp_pkt->buf, sizeof(udp_pkt->buf) );
    if( query == NULL ) {
	/* we MUST not send error messages */
	return 0;
    }
    
    debug("init answer from query ...\n");
    dns_init_answer(&answer,query);

    answer.from_mcast = udp_pkt->from_mcast;
    answer.from_local = udp_pkt->from_local;

    for( n = 0; n < query->nquest; n ++ ) {
	search_state s;
	info_init_search(&s,query->question[n].query_arg,query->question[n].query_type);

	while( info_search(&s) >0 ) {
	    answer_count ++;
	    dns_add_rr( &answer, s.data );
	}
    }

    if( answer_count > 0 ) {
      random_sleep();
      debug_dns("answer to client", &(answer.u.raw) );
      udp_send_dnsmsg( sock , &answer );
    }

    return 0;
}

/*****************************************************************************
 * Send a probe packet to the multicast address.
 *
 * The packet contains questions for all of out RR's plus all
 * the RR's in the authority section. 
 *****************************************************************************/
int probe( int sock ) {
    dns_t answer;
    search_state s;

    debug("Probe ...\n");
    dns_init(&answer);
    debug("init answer ...\n");
    dns_init_answer(&answer,NULL);
    answer.dst_port = -1;

    info_init_search(&s,NULL,T_ANY);
    while( info_search(&s) >0 ) {
        dns_add_qr( &answer, s.data->domain, s.data->type, C_IN );
    }


    info_init_search(&s,NULL,T_ANY);
    while( info_search(&s) >0 ) {
        dns_add_ar( &answer, s.data );
    }

    /* this is a query */
    answer.u.hdr.qr = 0;

    /* we MUST set the query-id to 0 */
    answer.u.hdr.id = 0;

    debug_dns("send probe ...\n", &(answer.u.raw));
    udp_send_dnsmsg( sock , &answer );

    return answer.u.hdr.id;

}

/*****************************************************************************
 * Send announcement packet to the multicast address.
 *
 * An announcement is simply an answer to all of our RR's
 *
 *****************************************************************************/
int announce( int sock ) {
    dns_t answer;
    search_state s;

    debug("Announce ...\n");
    dns_init(&answer);
    debug("init answer ...\n");
    dns_init_answer(&answer,NULL);
    answer.dst_port = -1;

    info_init_search(&s,NULL,T_ANY);
    while( info_search(&s) >0 ) {
        dns_add_rr( &answer, s.data );
	syslog(LOG_INFO,"claim name \"%s\", type %d\n" , s.data->domain , s.data->type );
    }

    /* this is an answer */
    answer.u.hdr.qr = 1;

    debug_dns("send announce ...\n", &(answer.u.raw));
    udp_send_dnsmsg( sock , &answer );

    return answer.u.hdr.id;

}

/*****************************************************************************
 * Add a local (bridged) query to our list of outstanding queries.
 *
 * We create a new dns_t structure fro the answer and add our local RR's
 * to this list (as we do it for mcast queries) but we do not answer this
 * query imediately. Instead, we send this query as a multicast query.
 *
 * When we get back some answers for the query, we will add the answer
 * records to the stored query. 
 *
 * After a timeout of about two seconds we answer the question to the
 * unicast socket.
 *****************************************************************************/
int store_query( int sock,  struct udp_packet *udp_pkt ) {

    int n = 0;
    int answer_count  = 0;
    dns_query * query = NULL;
    dns_t * answer;
    dns_t mcast_query;
    int oldest = 0;

    {
	ll_entry_t * el = ll_first(query_list);
	if( el ) {
	    oldest = ((stored_query *)el->data)->answer->timeout;
	}
    }
    
    answer = (dns_t *)malloc( sizeof(dns_t) );

    dns_init(answer);

    debug_dns("query from client", &(udp_pkt->buf));

    udp_copy_answer_address(answer,udp_pkt);

    answer->fd = sock;

    answer->timeout = time(NULL) + config.gather_delay;

    if( oldest == 0 ) { oldest = answer->timeout; }

    debug("decode query ...\n");
    query = dns_decode_query(udp_pkt->buf, sizeof(udp_pkt->buf) );

    if( query == NULL ) {
	/* 
	 * seems we got an error when decoding the query, send back
	 * an error. (this is unicast)
	 */
	answer->u.hdr.rcode = FORMERR;
	udp_send_dnsmsg( sock , answer );

	free(answer);

	return oldest;
    }
    
    debug("init answer from query ...\n");
    dns_init_answer(answer,query);
    answer->from_mcast = udp_pkt->from_mcast;
    answer->from_local = udp_pkt->from_local;

    if( ! config.allow_nonlocal ) {
        int respond_servfail = 0;

        for( n = 0; n < query->nquest; n ++ ) {
	    int matched = 0;   
	    int didx = 0;

	    for( didx = 0; config.also_local[didx] != NULL; didx ++ ) {
	        char * suffix = config.also_local[didx];
		int    suflen = strlen(suffix);
		int    qlen   = strlen(query->question[n].query_arg);
		char * end = NULL;

		if( qlen < suflen ) {
		    continue;
		}

		end = &query->question[n].query_arg[qlen - suflen];
		
		if ( strncasecmp( end , suffix, suflen ) == 0 ) {
		    matched = 1;
		}
	    }
	    if( ! matched ) {
		respond_servfail = 1;
	    }
	}

	if( respond_servfail ) {
	    debug("respond with SERVFAIL because query is not link-local\n");
	    answer->u.hdr.rcode = SERVFAIL;
	    udp_send_dnsmsg( sock , answer );
	    free(answer);
	    return oldest;
	}
    }


    for( n = 0; n < query->nquest; n ++ ) {
	search_state s;

	info_init_search(&s,query->question[n].query_arg,query->question[n].query_type);

	while( info_search(&s) >0 ) {
	    answer_count ++;
	    dns_add_rr( answer, s.data );
	}
    }

    {
	stored_query * sq = (stored_query *)malloc(sizeof(stored_query));
	dns_query * qp    = (dns_query *)malloc(sizeof(dns_query));

	memcpy(qp,query,sizeof(dns_query));

	sq->answer = answer;
	sq->query  = qp;

        ll_add(query_list,sq);
    }


    debug("init mcast query ...\n");
    dns_init(&mcast_query);
    dns_init_answer(&mcast_query,query);
    mcast_query.dst_port = -1;

    /* this is a query */
    mcast_query.u.hdr.qr = 0;

    debug_dns("send mcast query ...\n", &(mcast_query.u.raw));
    udp_send_dnsmsg( sock , &mcast_query);

    return oldest;
}

/*****************************************************************************
 * Answer outstanding queries.
 *
 *
 *****************************************************************************/
int answer_outstanding_queries() {

    ll_entry_t * el;
    int now;

    now = time(NULL);

    el = ll_first(query_list);
    while( el != NULL ) {
	ll_entry_t * next;
	dns_t     * answer;
	dns_query * query;

	next = ll_next(el);
	answer = ((stored_query *)el->data)->answer;
	query  = ((stored_query *)el->data)->query;

	if( answer->timeout <= now ) {
    	    if( ntohs(answer->u.hdr.ancount) > 0 ) {
        	debug_dns("answer to client", &(answer->u.raw) );
        	udp_send_dnsmsg( answer->fd, answer );
	    } else {
        	answer->u.hdr.rcode = NXDOMAIN;
        	udp_send_dnsmsg( answer->fd , answer );
	    }

	    free(answer);
	    free(query);
	    free(el->data);
	    ll_remove(el);
	} else {
	  /*
	   * Our list is kept in the order we've got the queries. Whenever
	   * we find a record that is not to be send now, we can return.
	   */
	   return answer->timeout ;
	}

	el = next;
    }

    return 0;
}


typedef struct {
    int	     found;
    char *   domain;
    int      type;
    int      class;
    int      rr_len;
    u_char * rr_data;
} _search_rr;

#ifndef min
#define min(a,b)  (a<b?a:b)
#endif

int dup_check_cb ( 
	      dns_sect section,
              char * domain,
              int type, int class, long ttl,
              int rr_len ,u_char * rr_data,
              void * user_data )
{

    _search_rr * s = (_search_rr *)user_data;

    if( section != IN_ANSWER_SECT ) { return 0; }

    if(              ( s->type   == type )                           &&
	             ( s->class  == class)                           &&
	             ( s->rr_len == rr_len)                          &&
	         ( strcmp(s->domain,domain) == 0 )                   &&
	( memcmp( s->rr_data, rr_data , min(s->rr_len,rr_len)) == 0 ))
    {
	s->found = 1;
        debug("dup rr %s, type %d\n",domain,type);
    }

    return 0;
}

int fill_answer_cb ( 
	      dns_sect section,
              char * domain,
              int type, int class, long ttl,
              int rr_len ,u_char * rr_data,
              void * user_data )
{

    int n = 0;
    dns_t * answer    = ((stored_query *)user_data)->answer;
    dns_query * query = ((stored_query *)user_data)->query;
    
    assert(query  != NULL );
    assert(answer != NULL );

    if( section != IN_ANSWER_SECT ) { return 0; }

    for( n = 0; n < query->nquest; n ++ ) {
	debug("question %d\n",n);
        if( ( (query->question[n].query_type == T_ANY) || 
	      (query->question[n].query_type == type ) ) &&
	    ( query->question[n].query_class == class ) &&
            ( strcmp(query->question[n].query_arg,domain) == 0) ) 
	{
	    _search_rr s;
	    memset(&s,0,sizeof(s));
	    s.found  = 0;
	    s.domain = domain;
	    s.type   = type;
	    s.class  = class;
	    s.rr_len = rr_len;
	    s.rr_data= rr_data;

	    dns_walk_buf( answer->u.raw, DNSDATASIZE, dup_check_cb , &s );

	    if( s.found == 0 ) {
	      debug("add rr for %s, type %d\n",domain,type);
	      dns_add_raw_rr(answer,domain,type,class,ttl,rr_len,rr_data);
	    } else {
	      debug("dup rr %s, type %d\n",domain,type);
	    }
	}
    }
    return 0;
}

/*****************************************************************************
 * Fill the RR's from the answer record we've got into the queries that 
 * may need them.
 *
 * We do no duplicat elimination.
 *
 *****************************************************************************/
void fill_outstanding_queries(int fd, struct udp_packet * pkt) {
    ll_entry_t * el;

    el = ll_first(query_list);

    debug("fill outstandig queries ...\n");
    for( el = ll_first(query_list); el != NULL; el = ll_next(el) ) {
	stored_query * sq = (stored_query *)el->data;
	assert(sq != NULL);
	dns_walk_packet( pkt , fill_answer_cb , sq );
    }
}

